"""indy-vdr Python wrapper library"""

from .bindings import set_config, set_protocol_version, version
from .error import VdrError, VdrErrorCode
from .ledger import LedgerType
from .pool import open_pool, Pool
from .request import Request

__all__ = [
    "open_pool",
    "set_config",
    "set_protocol_version",
    "version",
    "LedgerType",
    "Pool",
    "Request",
    "VdrError",
    "VdrErrorCode",
]



"""Methods for generating and working with pool ledger requests."""

from ctypes import byref, c_int32, c_int64, c_uint64
from datetime import datetime, date
from enum import IntEnum
from typing import Optional, Union

from .bindings import RequestHandle, do_call, encode_json, encode_str, lib_string
from .request import Request


class LedgerType(IntEnum):
    """An enum representing the sub-ledger indexes."""

    POOL = 0
    DOMAIN = 1
    CONFIG = 2

    @classmethod
    def from_value(cls, val: [int, str, "LedgerType"]) -> "LedgerType":
        """Initialize a `LedgerType` from an integer or string."""
        if isinstance(val, str):
            if val.isdigit():
                return cls(int(val))
            else:
                return cls[val.upper()]
        elif isinstance(val, int):
            return cls(val)
        elif isinstance(val, LedgerType):
            return val
        raise TypeError


def build_acceptance_mechanisms_request(
    submitter_did: str,
    aml: Union[str, bytes, dict],
    version: str,
    aml_context: str = None,
) -> Request:
    """
    Builds a SET_TXN_AUTHR_AGRMT_AML request.

    Request to add a new list of acceptance mechanisms for transaction author
    agreement. Acceptance Mechanism is a description of the ways how the user may
    accept a transaction author agreement.

    Args:
        submitter_did: Identifier (DID) of the transaction author as a base58-encoded
            string.
        aml: a set of new acceptance mechanisms:
            {
                "<acceptance mechanism label 1>": { description 1},
                "<acceptance mechanism label 2>": { description 2},
                ...
            }
        version: The version of the new acceptance mechanisms. (Note: unique on the
            Ledger)
        aml_context: (Optional) common context information about acceptance mechanisms
            (may be a URL to external resource)
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    aml_p = encode_str(aml) if isinstance(aml, (str, bytes)) else encode_json(aml)
    version_p = encode_str(version)
    aml_context_p = encode_str(aml_context)
    do_call(
        "indy_vdr_build_acceptance_mechanisms_request",
        did_p,
        aml_p,
        version_p,
        aml_context_p,
        byref(handle),
    )
    return Request(handle)


def build_attrib_request(
    submitter_did: Optional[str],
    target_did: str,
    xhash: Optional[str],
    raw: Optional[str],
    enc: Optional[str],
) -> str:
    """
    Builds an ATTRIB request.

    Request to add attribute to a NYM record.

    Args:
        submitter_did: Identifier (DID) of the transaction author as base58-encoded
            string.
        target_did: Target DID as base58-encoded string for 16 or 32 bit DID value.
        xhash: (Optional) Hash of attribute data.
        raw: (Optional) JSON, where key is attribute name and value is attribute value.
        enc: (Optional) Encrypted value attribute data.
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    target_p = encode_str(target_did)
    raw_p = encode_str(raw)
    hash_p = encode_str(xhash)
    enc_p = encode_str(enc)
    do_call(
        "indy_vdr_build_attrib_request",
        did_p,
        target_p,
        hash_p,
        raw_p,
        enc_p,
        byref(handle),
    )
    return Request(handle)


def build_cred_def_request(
    submitter_did: str, cred_def: Union[bytes, str, dict]
) -> Request:
    """
    Builds a CRED_DEF request to to add a credential definition to the ledger.

    In particular, this publishes the public key that the issuer created for
    issuing credentials against a particular schema.

    Args:
        submitter_did: Identifier (DID) of the transaction author as base58-encoded
            string
        cred_def: Credential definition.
            ```jsonc
            {
                "id": "<credential definition identifier>",
                "schemaId": "<schema identifier>",
                "type": "CL",
                    // type of the credential definition. CL is currently
                    // the only supported type
                "tag": "",
                    // allows to distinguish between credential definitions
                    // for the same issuer and schema
                "value": /* Dictionary with Credential Definition's data: */ {
                    "primary": "<primary credential public key>",
                    "revocation": /* Optional */ "<revocation credential public key>"
                },
                ver: Version of the CredDef json
            }```
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    cred_def_p = (
        encode_str(cred_def)
        if isinstance(cred_def, (str, bytes))
        else encode_json(cred_def)
    )
    do_call("indy_vdr_build_cred_def_request", did_p, cred_def_p, byref(handle))
    return Request(handle)


def build_custom_request(body: Union[str, bytes, dict]) -> Request:
    handle = RequestHandle()
    body_p = encode_str(body) if isinstance(body, (str, bytes)) else encode_json(body)
    do_call("indy_vdr_build_custom_request", body_p, byref(handle))
    return Request(handle)


def build_disable_all_txn_author_agreements_request(submitter_did: str) -> Request:
    """
    Builds a DISABLE_ALL_TXN_AUTHR_AGRMTS request.

    Used to disable all Transaction Author Agreements on the ledger.

    Args:
        submitter_did: Identifier (DID) of the transaction author as base58-encoded
            string.
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    do_call(
        "indy_vdr_build_disable_all_txn_author_agreements_request", did_p, byref(handle)
    )
    return Request(handle)


def build_get_acceptance_mechanisms_request(
    submitter_did: str = None, timestamp: int = None, version: str = None
) -> Request:
    """
    Builds a GET_TXN_AUTHR_AGRMT_AML request.

    Request to get a list of acceptance mechanisms from the ledger valid for specified
    time, or the latest one.

    Args:
        submitter_did: (Optional) DID of the read request sender (if not provided, then
            the default Libindy DID will be used)
        timestamp: (Optional) time to get an active acceptance mechanisms. The latest
            one will be returned for the empty timestamp
        version: (Optional) version of acceptance mechanisms

    NOTE: timestamp and version cannot be specified together.
    """

    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    timestamp_c = c_int64(timestamp if timestamp is not None else -1)
    version_p = encode_str(version)
    do_call(
        "indy_vdr_build_get_acceptance_mechanisms_request",
        did_p,
        timestamp_c,
        version_p,
        byref(handle),
    )
    return Request(handle)


def build_get_attrib_request(
    submitter_did: Optional[str],
    target_did: str,
    raw: Optional[str],
    xhash: Optional[str],
    enc: Optional[str],
) -> str:
    """
    Builds a GET_ATTRIB request.

    Request to get information about an Attribute for the specified DID.

    Args:
        submitter_did: (Optional) DID of the read request sender (if not provided, then
            the default Libindy DID will be used).
        target_did: Target DID as base58-encoded string for 16 or 32 bit DID value.
        xhash: (Optional) Requested attribute name.
        raw: (Optional) Requested attribute hash.
        enc: (Optional) Requested attribute encrypted value.
    """

    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    target_p = encode_str(target_did)
    raw_p = encode_str(raw)
    hash_p = encode_str(xhash)
    enc_p = encode_str(enc)
    do_call(
        "indy_vdr_build_get_attrib_request",
        did_p,
        target_p,
        raw_p,
        hash_p,
        enc_p,
        byref(handle),
    )
    return Request(handle)


def build_get_cred_def_request(
    submitter_did: Optional[str], cred_def_id: str
) -> Request:
    """
    Builds a GET_CRED_DEF request to fetch a credential definition by ID.

    Args:
        submitter_did: (Optional) DID of the read request sender. If not provided
            then the default Libindy DID will be used
        cred_def_id: ID of the corresponding credential definition on the ledger
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    cred_def_id_p = encode_str(cred_def_id)
    do_call("indy_vdr_build_get_cred_def_request", did_p, cred_def_id_p, byref(handle))
    return Request(handle)


def build_get_nym_request(submitter_did: Optional[str], dest: str) -> Request:
    """
    Builds a GET_NYM request to get information about a DID (NYM).

    Args:
        submitter_did: (Optional) DID of the read request sender. If not provided
            then the default Libindy DID will be use)
        target_did: Target DID as base58-encoded string for 16 or 32 bit DID value
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    dest_p = encode_str(dest)
    do_call("indy_vdr_build_get_nym_request", did_p, dest_p, byref(handle))
    return Request(handle)


def build_get_revoc_reg_def_request(
    submitter_did: Optional[str], revoc_reg_id: str
) -> Request:
    """
    Builds a GET_REVOC_REG_DEF request.

    Request to get the revocation registry definition for a given revocation
    registry ID.

    Args:
        submitter_did: (Optional) DID of the read request sender. If not provided
            then the default Libindy DID will be used.
        revoc_reg_id: ID of the corresponding revocation registry definition
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    rev_id_p = encode_str(revoc_reg_id)
    do_call("indy_vdr_build_get_revoc_reg_def_request", did_p, rev_id_p, byref(handle))
    return Request(handle)


def build_get_revoc_reg_request(
    submitter_did: Optional[str], revoc_reg_id: str, timestamp: int
) -> Request:
    """
    Builds a GET_REVOC_REG request.

    Request to get the accumulated state of the revocation registry by ID. The state
    is defined by the given timestamp.

    Args:
        submitter_did: (Optional) DID of the read request sender. If not provided
            then the default Libindy DID will be used
        revoc_reg_id: ID of the corresponding revocation registry definition
        timestamp: Requested time represented as a total number of seconds since the
            Unix epoch
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    rev_id_p = encode_str(revoc_reg_id)
    timestamp_c = c_int64(timestamp)
    do_call(
        "indy_vdr_build_get_revoc_reg_request",
        did_p,
        rev_id_p,
        timestamp_c,
        byref(handle),
    )
    return Request(handle)


def build_get_revoc_reg_delta_request(
    submitter_did: Optional[str], revoc_reg_id: str, from_ts: Optional[int], to_ts: int
) -> Request:
    """
    Builds a GET_REVOC_REG_DELTA request.

    Request to get the delta of the accumulated state of the revocation registry
    identified by `revoc_reg_id`. The delta is defined by from and to timestamp fields.
    If from is not specified, then the whole state until `to` will be returned.

    Args:
        submitter_did: (Optional) DID of the read request sender. If not provided
            then the default Libindy DID will be used
        revoc_reg_id: ID of the corresponding revocation registry definition
        from_ts: Requested time represented as a total number of seconds from Unix epoch
        to_ts: Requested time represented as a total number of seconds from Unix epoch
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    rev_id_p = encode_str(revoc_reg_id)
    from_c = c_int64(from_ts if from_ts is not None else -1)
    to_c = c_int64(to_ts)
    do_call(
        "indy_vdr_build_get_revoc_reg_delta_request",
        did_p,
        rev_id_p,
        from_c,
        to_c,
        byref(handle),
    )
    return Request(handle)


def build_get_schema_request(submitter_did: Optional[str], schema_id: str) -> Request:
    """
    Builds a GET_SCHEMA request to fetch a credential schema by ID.

    Args:
        submitter_did: (Optional) DID of the read request sender. If not provided
            then the default Libindy DID will be used
        schema_id: ID of the corresponding schema on the ledger
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    schema_id_p = encode_str(schema_id)
    do_call("indy_vdr_build_get_schema_request", did_p, schema_id_p, byref(handle))
    return Request(handle)


def build_get_txn_author_agreement_request(
    submitter_did: str = None, data: Union[bytes, str, dict] = None
) -> Request:
    """
    Builds a GET_TXN_AUTHR_AGRMT request.

    Used to get a specific Transaction Author Agreement from the ledger.

    Args:
        submitter_did: (Optional) DID of the read request sender. If not provided
            then the default Libindy DID will be used
        data: (Optional) specifies conditions for getting a specific TAA
            Contains 3 mutually exclusive optional fields:
            {
                hash: Optional<str> - hash of requested TAA,
                version: Optional<str> - version of requested TAA.
                timestamp: Optional<i64> - ledger will return TAA valid at requested
                    timestamp.
            }
            Null data or empty JSON are acceptable here. In this case, ledger will
            return the latest version of the TAA.
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    data_p = (
        encode_str(data)
        if isinstance(data, (str, bytes)) or data is None
        else encode_json(data)
    )
    do_call(
        "indy_vdr_build_get_txn_author_agreement_request", did_p, data_p, byref(handle)
    )
    return Request(handle)


def build_get_txn_request(
    submitter_did: Optional[str],
    ledger_type: Optional[Union[int, str, LedgerType]],
    seq_no: int,
) -> Request:
    """
    Builds a GET_TXN request to get any transaction by its sequence number.

    Args:
        submitter_did: (Optional) DID of the read request sender. If not provided
            then the default Libindy DID will be used
        ledger_type: (Optional) type of the ledger the requested transaction belongs to
            Pass a `LedgerType` instance for known values
        seq_no: requested transaction sequence number as it's stored on the ledger
    """

    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    ledger_type = (
        LedgerType.from_value(ledger_type)
        if ledger_type is not None
        else LedgerType.DOMAIN
    )
    do_call(
        "indy_vdr_build_get_txn_request",
        did_p,
        c_int32(ledger_type),
        c_int32(seq_no),
        byref(handle),
    )
    return Request(handle)


def build_get_validator_info_request(submitter_did: str) -> Request:
    """
    Builds a GET_VALIDATOR_INFO request.

    Args:
        submitter_did: DID of the request sender
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    do_call("indy_vdr_build_get_validator_info_request", did_p, byref(handle))
    return Request(handle)


def build_nym_request(
    submitter_did: str,
    dest: str,
    verkey: str = None,
    alias: str = None,
    role: str = None,
) -> Request:
    """
    Builds a NYM request to create new DID on the ledger.

    Args:
        submitter_did: Identifier (DID) of the transaction author as base58-encoded
            string
        target_did: Target DID as base58-encoded string for 16 or 32 bit DID value
        verkey: (Optional) Target identity verification key as base58-encoded string
        alias: (Optional) The NYM's alias.
        role: (Optional) Role of a user NYM record:
            null (common USER)
            TRUSTEE
            STEWARD
            TRUST_ANCHOR
            ENDORSER - equal to TRUST_ANCHOR that will be removed soon
            NETWORK_MONITOR
            empty string to reset role
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    dest_p = encode_str(dest)
    verkey_p = encode_str(verkey) if verkey else None
    alias_p = encode_str(alias) if alias else None
    role_p = encode_str(role) if role else None
    do_call(
        "indy_vdr_build_nym_request",
        did_p,
        dest_p,
        verkey_p,
        alias_p,
        role_p,
        byref(handle),
    )
    return Request(handle)


def build_revoc_reg_def_request(
    submitter_did: str, revoc_reg_def: Union[bytes, str, dict]
) -> Request:
    """
    Builds a REVOC_REG_DEF request.

    Request to add the definition of revocation registry to an existing
    credential definition.

    Args:
        submitter_did: Identifier (DID) of the transaction author as base58-encoded
            string
        revoc_reg_def: Revocation Registry data:
            ```jsonc
            {
                "id": "<revocation registry identifier>",
                "revocDefType": "CL_ACCUM",
                    // revocation registry type (only CL_ACCUM is supported for now)
                "tag": "", // Unique descriptive ID of the registry definition
                "credDefId": "<credential definition ID>",
                "value": /* Registry-specific data */ {
                    "issuanceType": "ISSUANCE_BY_DEFAULT",
                        // Type of issuance: ISSUANCE_BY_DEFAULT or ISSUANCE_ON_DEMAND
                    "maxCredNum": 10000,
                        // Maximum number of credentials the Registry can serve.
                    "tailsHash": "<sha256 hash of tails file in base58>",
                    "tailsLocation": "<URL or path for the tails file>",
                    "publicKeys": { /* <public_keys> */ } // registry's public keys
                },
                "ver": "<version of revocation registry definition json>"
            }```
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    revoc_reg_def_p = (
        encode_str(revoc_reg_def)
        if isinstance(revoc_reg_def, (str, bytes))
        else encode_json(revoc_reg_def)
    )
    do_call(
        "indy_vdr_build_revoc_reg_def_request", did_p, revoc_reg_def_p, byref(handle)
    )
    return Request(handle)


def build_revoc_reg_entry_request(
    submitter_did: str,
    revoc_reg_def_id: str,
    revoc_reg_def_type: str,
    entry: Union[bytes, str, dict],
) -> Request:
    """
    Builds a REVOC_REG_ENTRY request.

    Request to add the revocation registry entry containing the new accumulator
    value and issued/revoked indices. This is just a delta of indices, not the
    whole list. It can be sent each time a new credential is issued/revoked.

    Args:
        submitter_did: Identifier (DID) of the transaction author as base58-encoded
            string
        entry: Registry-specific data:
            ```jsonc
            {
                "value": {
                    "prevAccum": "<previous accumulator value>",
                    "accum": "<current accumulator value>",
                    "issued": [], // array<number> - an array of issued indices
                    "revoked": [] // array<number> an array of revoked indices
                },
                "ver": "<version of the revocation registry entry json>"
            }```
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    def_id_p = encode_str(revoc_reg_def_id)
    def_type_p = encode_str(revoc_reg_def_type)
    entry_p = (
        encode_str(entry) if isinstance(entry, (str, bytes)) else encode_json(entry)
    )
    do_call(
        "indy_vdr_build_revoc_reg_entry_request",
        did_p,
        def_id_p,
        def_type_p,
        entry_p,
        byref(handle),
    )
    return Request(handle)


def build_schema_request(
    submitter_did: str, schema: Union[bytes, str, dict]
) -> Request:
    """
    Builds a SCHEMA request to to add a credential schema to the ledger.

    Args:
        submitter_did: Identifier (DID) of the transaction author as base58-encoded
            string
        schema: Credential schema:
            ```jsonc
            {
                "id": "<identifier of schema>",
                "attrNames": "<array of attribute name strings (the number of attributes
                    should be less or equal than 125)>",
                "name": "<schema's name string>",
                "version": "<schema's version string>",
                "ver": "<version of the schema json>"
            }```
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    schema_p = (
        encode_str(schema) if isinstance(schema, (str, bytes)) else encode_json(schema)
    )
    do_call("indy_vdr_build_schema_request", did_p, schema_p, byref(handle))
    return Request(handle)


def build_txn_author_agreement_request(
    submitter_did: str,
    text: Optional[str],
    version: str,
    ratification_ts: int = None,
    retirement_ts: int = None,
) -> Request:
    """
    Builds a TXN_AUTHR_AGRMT request.

    Used to add a new version of the Transaction Author Agreement to the ledger.

    Args:
        submitter_did: Identifier (DID) of the transaction author as base58-encoded
            string.
        text: (Optional) the content of the TAA. Mandatory in case of adding a new TAA.
            An existing TAA text can not be changed.
            For Indy Node version <= 1.12.0:
                Use empty string to reset TAA on the ledger
            For Indy Node version > 1.12.0:
                Should be omitted in case of updating an existing TAA (setting
                `retirement_ts`)
        version: the version of the TAA (a unique UTF-8 string)
        ratification_ts: (Optional) the date (timestamp) of TAA ratification by
            network government.
            For Indy Node version <= 1.12.0:
                Must be omitted
            For Indy Node version > 1.12.0:
                Must be specified in case of adding a new TAA
                Can be omitted in case of updating an existing TAA
        retirement_ts: (Optional) the date (timestamp) of TAA retirement
            For Indy Node version <= 1.12.0:
                Must be omitted
            For Indy Node version > 1.12.0:
                Must be omitted in case of adding a new (latest) TAA.
                Should be used for updating (deactivating) non-latest TAA on the
                ledger.

    Note: Use `build_disable_all_txn_author_agreements_request` to disable all TAAs
    on the ledger.
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    text_p = encode_str(text)
    version_p = encode_str(version)
    ratification_ts_c = c_int64(ratification_ts if ratification_ts is not None else -1)
    retirement_ts_c = c_int64(retirement_ts if retirement_ts is not None else -1)
    do_call(
        "indy_vdr_build_txn_author_agreement_request",
        did_p,
        text_p,
        version_p,
        ratification_ts_c,
        retirement_ts_c,
        byref(handle),
    )
    return Request(handle)


def prepare_txn_author_agreement_acceptance(
    text: Optional[str],
    version: Optional[str],
    taa_digest: Optional[str],
    mechanism: str,
    accepted_time: int = None,
) -> str:
    """
    Prepare transaction author agreement acceptance data.

    This function generates transaction author agreement data which can be appended
    to a specific request using `Request.set_txn_author_agreement_acceptance`, which
    must be used if there is any transaction author agreement set on the ledger.

    This function may calculate hash by itself or consume it as a parameter. If all
    text, version and taa_digest parameters are specified, an integrity check will be
    performed on them.

    Args:
        text: (Optional) raw TAA text from the ledger
        version: (Optional) raw TAA version from the ledger
        taa_digest: (Optional) digest of the TAA text and version.
            Digest is sha256 hash calculated on the concatenated strings:
            `version || text`. This parameter is required if text and version
            parameters are omitted
        mechanism: mechanism from the ledger Acceptance Mechanisms List indicating how
            the user has accepted the TAA
        time: UTC timestamp representing when user has accepted the TAA. Note that the
            time portion will be discarded to protect privacy
    """

    text_p = encode_str(text)
    version_p = encode_str(version)
    taa_digest_p = encode_str(taa_digest)
    mechanism_p = encode_str(mechanism)
    if not accepted_time:
        # rough timestamp
        accepted_time = int(
            datetime.combine(date.today(), datetime.min.time()).timestamp()
        )
    result = lib_string()
    do_call(
        "indy_vdr_prepare_txn_author_agreement_acceptance",
        text_p,
        version_p,
        taa_digest_p,
        mechanism_p,
        c_uint64(accepted_time),
        byref(result),
    )
    return result.value.decode("utf-8")


def build_rich_schema_request(
    submitter_did: str,
    rs_id: Union[bytes, str],
    rs_content: Union[bytes, str, dict],
    rs_name: Union[bytes, str],
    rs_version: Union[bytes, str],
    rs_type: Union[bytes, str],
    ver: Union[bytes, str],
) -> Request:
    """
    Builds a RICH_SCHEMA request to add it to the ledger.

    Args:
        submitter_did: Identifier (DID) of the transaction author as a base58-encoded
            string
        rs_id: identifier of the rich schema
        rs_content: JSON-LD string object
        rs_name: rich schema name
        rs_version: rich schema version
        rs_type: type constant as string, one of
            `ctx`, `sch`, `map`, `enc`, `cdf`, `pdf`
        ver: version as string
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    rs_id = encode_str(rs_id)
    rs_content = (
        encode_str(rs_content)
        if isinstance(rs_content, (str, bytes))
        else encode_json(rs_content)
    )
    rs_name = encode_str(rs_name)
    rs_version = encode_str(rs_version)
    rs_type = encode_str(rs_type)
    ver = encode_str(ver)
    do_call(
        "indy_vdr_build_rich_schema_request",
        did_p,
        rs_id,
        rs_content,
        rs_name,
        rs_version,
        rs_type,
        ver,
        byref(handle),
    )
    return Request(handle)


def build_get_rich_schema_object_by_id_request(
    submitter_did: str, rs_id: Union[bytes, str, dict]
) -> Request:
    """
    Builds a GET_RICH_SCHEMA_BY_ID request.

    Used to fetch a RICH_SCHEMA from the ledger using RICH_SCHEMA_ID.

    Args:
        submitter_did: Identifier (DID) of the transaction author as base58-encoded
            string
        rs_id: DID-string like object which represents id of requested RICH_SCHEMA
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    rs_id = encode_str(rs_id) if isinstance(rs_id, (str, bytes)) else encode_json(rs_id)
    do_call(
        "indy_vdr_build_get_schema_object_by_id_request", did_p, rs_id, byref(handle)
    )
    do_call(
        "indy_vdr_build_get_rich_schema_object_by_id_request",
        did_p,
        rs_id,
        byref(handle),
    )
    return Request(handle)


def build_get_rich_schema_object_by_metadata_request(
    submitter_did: str,
    rs_type: Union[bytes, str],
    rs_name: Union[bytes, str],
    rs_version: Union[bytes, str],
) -> Request:
    """
    Builds a GET_RICH_SCHEMA_BY_METADATA request.

    Used to fetch a RICH_SCHEMA from the ledger using the RICH_SCHEMA's metadata.

    Args:
        submitter_did: Identifier (DID) of the transaction author as base58-encoded
            string
        rs_type: Rich Schema object's type enum
        rs_name: Rich Schema object's name,
        rs_version: Rich Schema object's version,
    """
    handle = RequestHandle()
    did_p = encode_str(submitter_did)
    rs_type = encode_str(rs_type)
    rs_name = encode_str(rs_name)
    rs_version = encode_str(rs_version)
    do_call(
        "indy_vdr_build_get_schema_object_by_metadata_request",
        did_p,
        rs_type,
        rs_name,
        rs_version,
        byref(handle),
    )
    do_call(
        "indy_vdr_build_get_rich_schema_object_by_metadata_request",
        did_p,
        rs_type,
        rs_name,
        rs_version,
        byref(handle),
    )
    return Request(handle)



"""Error classes."""

from enum import IntEnum


class VdrErrorCode(IntEnum):
    SUCCESS = 0
    CONFIG = 1
    CONNECTION = 2
    FILESYSTEM = 3
    INPUT = 4
    RESOURCE = 5
    UNAVAILABLE = 6
    UNEXPECTED = 7
    POOL_NO_CONSENSUS = 30
    POOL_REQUEST_FAILED = 31
    POOL_TIMEOUT = 32
    WRAPPER = 99


class VdrError(Exception):
    def __init__(self, code: VdrErrorCode, message: str, extra: str = None):
        super().__init__(message)
        self.code = code
        self.extra = extra



"""Handling of prepared ledger requests."""

from typing import Union

from . import bindings
from .error import VdrError, VdrErrorCode


class Request:
    """A prepared ledger request."""

    def __init__(self, handle: bindings.RequestHandle):
        """Initialize the `Request` instance."""
        self.handle = handle

    @property
    def body(self) -> str:
        """Get the body of the prepared request as a JSON string."""
        if not self.handle:
            raise VdrError(VdrErrorCode.WRAPPER, "no request handle")
        return bindings.request_get_body(self.handle)

    def free(self):
        """Release the prepared request."""
        if hasattr(self, "handle") and self.handle:
            bindings.request_free(self.handle)
            self.handle = None

    @property
    def signature_input(self) -> bytes:
        """Get the signature input of this prepared request.

        This value should be passed to the signer in order to generate an
        appropriate signature for the request.

        Returns:
            A bytes instance with the canonical representation of the request.
        """
        if not self.handle:
            raise VdrError(VdrErrorCode.WRAPPER, "no request handle")
        return bindings.request_get_signature_input(self.handle)

    def set_endorser(self, endorser: str):
        """
        Set the endorser property of an existing request.

        When a transaction is expected to be sent to the ledger by an endorser, this
        property must be set in order to generate a valid signature.
        Note: Both Transaction Author and Endorser must sign the output request.
        More about Transaction Endorser:
            https://github.com/hyperledger/indy-node/blob/master/design/transaction_endorser.md
            https://github.com/hyperledger/indy-sdk/blob/master/docs/configuration.md

        Args:
            endorser_did: DID of the Endorser that will submit the transaction.
                The Endorser's DID must be present on the ledger.
        """
        if not self.handle:
            raise VdrError(VdrErrorCode.WRAPPER, "no request handle")
        bindings.request_set_endorser(self.handle, endorser)

    def set_signature(self, signature: bytes):
        """Set the signature on a prepared request.

        Args:
            signature: A bytes instance with the ed25519 signature
                generated over `request.signature_input`
        """
        if not self.handle:
            raise VdrError(VdrErrorCode.WRAPPER, "no request handle")
        bindings.request_set_signature(self.handle, signature)

    def set_txn_author_agreement_acceptance(self, acceptance: Union[str, dict]):
        """Set the TAA acceptance on a prepared request.

        Args:
            acceptance: Normally be the output of
                `ledger.prepare_txn_agreement_acceptance`, which may be
                cached for the current session
        """
        if not self.handle:
            raise VdrError(VdrErrorCode.WRAPPER, "no request handle")
        bindings.request_set_txn_author_agreement_acceptance(self.handle, acceptance)

    def __del__(self):
        """Release the pool instance."""
        self.free()

    def __repr__(self) -> str:
        """Format the pool instance as a debug string."""
        if self.handle:
            status = self.handle
        else:
            status = "freed"
        return f"{self.__class__.__name__}({status})"



"""Handling of ledger pool instances."""

import json
from datetime import datetime
from typing import Mapping, Sequence, Union

from . import bindings
from .error import VdrError, VdrErrorCode
from .ledger import Request, build_custom_request


class Pool:
    """An opened ledger pool instance."""

    def __init__(self, handle: bindings.PoolHandle):
        """Initialize the pool instance."""
        self.handle = handle
        self.last_refresh: datetime = None
        self.last_status: dict = None

    def close(self):
        """Close and free the pool instance."""
        if hasattr(self, "handle") and self.handle:
            bindings.pool_close(self.handle)
            self.handle = None

    async def get_status(self) -> dict:
        """Get the current status of the pool instance."""
        if not self.handle:
            raise VdrError(VdrErrorCode.WRAPPER, "pool is closed")
        result = await bindings.pool_get_status(self.handle)
        self.last_status = json.loads(result)
        return result

    async def get_transactions(self) -> str:
        """Get the current pool transactions of the pool instance."""
        if not self.handle:
            raise VdrError(VdrErrorCode.WRAPPER, "pool is closed")
        return await bindings.pool_get_transactions(self.handle)

    async def get_verifiers(self) -> dict:
        """Get the current set of active verifiers for the pool instance."""
        if not self.handle:
            raise VdrError(VdrErrorCode.WRAPPER, "pool is closed")
        return json.loads(await bindings.pool_get_verifiers(self.handle))

    @property
    def last_refresh_seconds(self) -> float:
        """Get the number of seconds since the last verifier pool refresh."""
        return (
            (datetime.now() - self.last_refresh).total_seconds()
            if self.last_refresh
            else None
        )

    async def refresh(self) -> dict:
        """Check the verifier pool and load any new pool transactions."""
        await bindings.pool_refresh(self.handle)
        result = await bindings.pool_get_status(self.handle)
        self.last_status = json.loads(result)
        self.last_refresh = datetime.now()
        return self.last_status

    async def submit_action(
        self,
        request: Union[str, bytes, dict, Request],
        node_aliases: Sequence[str] = None,
        timeout: int = None,
    ) -> dict:
        """Submit a pool action to all verifier nodes.

        The following requests are sent as actions:
            GET_VALIDATOR_INFO
            POOL_RESTART

        Args:
            request: May be a prepared `Request` instance, a JSON string or bytes
                instance, or a dict representing a new custom ledger request

        Returns:
            A dict with the node aliases as keys and the node's responses
            as values
        """
        if not isinstance(request, Request):
            request = build_custom_request(request)
        if not self.handle:
            raise VdrError(VdrErrorCode.WRAPPER, "pool is closed")
        if not request.handle:
            raise VdrError(VdrErrorCode.WRAPPER, "no request handle")
        fut = bindings.pool_submit_action(
            self.handle, request.handle, node_aliases, timeout
        )
        request.handle = None  # request has been removed
        result = await fut
        return json.loads(result)

    async def submit_request(self, request: Union[str, bytes, dict, Request]) -> dict:
        """Submit a ledger request.

        Args:
            request: May be a prepared `Request` instance, a JSON string or bytes
                instance, or a dict representing a new custom ledger request

        Returns:
            A dict representing the decoded JSON response
        """
        if not isinstance(request, Request):
            request = build_custom_request(request)
        if not self.handle:
            raise VdrError(VdrErrorCode.WRAPPER, "pool is closed")
        if not request.handle:
            raise VdrError(VdrErrorCode.WRAPPER, "no request handle")
        fut = bindings.pool_submit_request(self.handle, request.handle)
        request.handle = None  # request has been removed
        result = await fut
        # FIXME improve handling of bad request
        return json.loads(result)["result"]

    def __del__(self):
        """Close the pool instance when there are no more references to this object."""
        self.close()

    def __repr__(self) -> str:
        """Format the pool instance as a debug string."""
        if self.handle:
            last_refresh = self.last_refresh_seconds
            last_refresh = int(last_refresh) if last_refresh is not None else None
            mt_size = self.last_status and self.last_status.get("mt_size")
            node_count = len(self.last_status and self.last_status.get("nodes") or ())
            status = (
                f"({self.handle.value}, count={node_count}, "
                f"last_refresh={last_refresh}, mt_size={mt_size})"
            )
        else:
            status = "(closed)"
        return f"{self.__class__.__name__}{status}"


async def open_pool(
    transactions_path: str = None,
    transactions: str = None,
    node_weights: Mapping[str, float] = None,
    no_refresh: bool = False,
) -> Pool:
    """Create a new ledger pool instance.

    Either `transactions` or `transactions_path` must be specified, but not both.

    Args:
        transactions_path: The path to a genesis transactions file
        transactions: A JSON string representing the genesis transactions
        node_weights: A dict with node aliases as the keys and priority weights
            as the values. The default weight is 1.0, so higher weights give the
            node a higher probability of being selected. A weight of zero means the
            node will never be selected.
        no_refresh: Disable the initial verifier pool refresh

    Returns:
        A new `Pool` instance which may be used to submit ledger requests
    """
    if not (bool(transactions_path) ^ bool(transactions)):
        raise VdrError(
            VdrErrorCode.WRAPPER,
            "Must provide one of transactions or transactions_path",
        )
    params = {
        "transactions": transactions,
        "transactions_path": transactions_path,
        "node_weights": node_weights,
    }
    pool = Pool(bindings.pool_create(params))
    if not no_refresh:
        await pool.refresh()
    return pool



"""Low-level interaction with the indy_vdr library."""

import asyncio
import json
import logging
import os
import sys
from ctypes import CDLL, CFUNCTYPE, byref, c_char_p, c_size_t, c_void_p, c_int32
from ctypes.util import find_library
from typing import Optional, Sequence, Union

from .error import VdrError, VdrErrorCode


CALLBACKS = {}
LIB: CDLL = None
LOGGER = logging.getLogger(__name__)


class PoolHandle(c_size_t):
    """Index of an active Pool instance."""

    def __repr__(self) -> str:
        """Format pool handle as a string."""
        return f"{self.__class__.__name__}({self.value})"


class RequestHandle(c_size_t):
    """Index of an active Request instance."""

    def __repr__(self) -> str:
        """Format request handle as a string."""
        return f"{self.__class__.__name__}({self.value})"


class lib_string(c_char_p):
    """A string allocated by the library."""

    @classmethod
    def from_param(cls):
        """Returns the type ctypes should use for loading the result."""
        return c_void_p

    def __bytes__(self):
        """Convert to bytes."""
        return self.value

    def __str__(self):
        """Convert to str."""
        return self.value.decode("utf-8")

    def __del__(self):
        """Call the string destructor when this instance is released."""
        get_library().indy_vdr_string_free(self)


def get_library() -> CDLL:
    """Return the CDLL instance, loading it if necessary."""
    global LIB
    if LIB is None:
        LIB = _load_library("indy_vdr")
        do_call("indy_vdr_set_default_logger")
    return LIB


def _load_library(lib_name: str) -> CDLL:
    """Load the CDLL library.

    The python module directory is searched first, followed by the usual
    library resolution for the current system.
    """
    lib_prefix_mapping = {"win32": ""}
    lib_suffix_mapping = {"darwin": ".dylib", "win32": ".dll"}
    try:
        os_name = sys.platform
        lib_prefix = lib_prefix_mapping.get(os_name, "lib")
        lib_suffix = lib_suffix_mapping.get(os_name, ".so")
        lib_path = os.path.join(
            os.path.dirname(__file__), f"{lib_prefix}{lib_name}{lib_suffix}"
        )
        return CDLL(lib_path)
    except KeyError:
        LOGGER.debug("Unknown platform for shared library")
    except OSError:
        LOGGER.warning("Library not loaded from python package")

    lib_path = find_library(lib_name)
    if not lib_path:
        raise VdrError(VdrErrorCode.WRAPPER, f"Error loading library: {lib_name}")
    try:
        return CDLL(lib_path)
    except OSError as e:
        raise VdrError(
            VdrErrorCode.WRAPPER, f"Error loading library: {lib_name}"
        ) from e


def _fulfill_future(fut: asyncio.Future, result, err: Exception = None):
    """Resolve a callback future given the result and exception, if any."""
    if fut.cancelled():
        LOGGER.debug("callback previously cancelled")
    elif err:
        fut.set_exception(err)
    else:
        fut.set_result(result)


def _create_callback(cb_type: CFUNCTYPE, fut: asyncio.Future, post_process=None):
    """Create a callback to handle the response from an async library method."""

    def _cb(id: int, err: int, result=None):
        """Callback function passed to the CFUNCTYPE for invocation."""
        if post_process:
            result = post_process(result)
        exc = get_current_error() if err else None
        try:
            (loop, _cb) = CALLBACKS.pop(fut)
        except KeyError:
            LOGGER.debug("callback already fulfilled")
            return
        loop.call_soon_threadsafe(lambda: _fulfill_future(fut, result, exc))

    res = cb_type(_cb)
    return res


def do_call(fn_name, *args):
    """Perform a synchronous library function call."""
    lib_fn = getattr(get_library(), fn_name)
    result = lib_fn(*args)
    if result:
        raise get_current_error(True)


def do_call_async(fn_name, *args, return_type=None, post_process=None):
    """Perform an asynchronous library function call."""
    lib_fn = getattr(get_library(), fn_name)
    loop = asyncio.get_event_loop()
    fut = loop.create_future()
    cf_args = [None, c_size_t, c_size_t]
    if return_type:
        cf_args.append(return_type)
    cb_type = CFUNCTYPE(*cf_args)  # could be cached
    cb_res = _create_callback(cb_type, fut, post_process)
    # keep a reference to the callback function to avoid it being freed
    CALLBACKS[fut] = (loop, cb_res)
    result = lib_fn(*args, cb_res, c_size_t(0))  # not making use of callback ID
    if result:
        # callback will not be executed
        del CALLBACKS[fut]
        fut.set_exception(get_current_error())
    return fut


def encode_json(arg) -> c_char_p:
    """Encode an input argument as JSON."""
    return encode_str(json.dumps(arg))


def encode_str(arg: Optional[Union[str, bytes]]) -> c_char_p:
    """Encode an optional input argument as a string.

    Returns: None if the argument is None, otherwise the value encoded utf-8.
    """
    if arg is None:
        return None
    if isinstance(arg, bytes):
        return c_char_p(arg)
    return c_char_p(arg.encode("utf-8"))


def get_current_error(expect: bool = False) -> VdrError:
    """Get the error result from the previous failed API method.

    Args:
        expect: Return a default error message if none is found
    """
    err_json = lib_string()
    if not get_library().indy_vdr_get_current_error(byref(err_json)):
        try:
            msg = json.loads(err_json.value)
        except json.JSONDecodeError:
            LOGGER.warning("JSON decode error for indy_vdr_get_current_error")
            msg = None
        if msg and "message" in msg and "code" in msg:
            return VdrError(VdrErrorCode(msg["code"]), msg["message"], msg.get("extra"))
        if not expect:
            return None
    return VdrError(VdrError.WRAPPER, "Unknown error")


def pool_create(params: Union[str, bytes, dict]) -> PoolHandle:
    """Create a new pool instance.

    Args:
        params: A JSON-encoded str or bytes instance or a dict representing the
            pool creation parameters. See `pool.open_pool` for the parameters.
    """
    handle = PoolHandle()
    params_p = (
        encode_str(params) if isinstance(params, (str, bytes)) else encode_json(params)
    )
    do_call("indy_vdr_pool_create", params_p, byref(handle))
    return handle


def pool_get_status(pool_handle: PoolHandle) -> asyncio.Future:
    """Get the status of an opened pool instance."""
    return do_call_async(
        "indy_vdr_pool_get_status",
        pool_handle,
        return_type=lib_string,
        post_process=str,
    )


def pool_refresh(pool_handle: PoolHandle) -> asyncio.Future:
    """Fetch the latest transactions for the ledger's verifier pool."""
    return do_call_async("indy_vdr_pool_refresh", pool_handle)


def pool_submit_action(
    pool_handle: PoolHandle,
    request_handle: RequestHandle,
    node_aliases: Sequence[str] = None,
    timeout: int = None,
) -> asyncio.Future:
    """Publishes a prepared pool action request message to the validator pool."""
    nodes_p = encode_json(node_aliases) if node_aliases else c_void_p()
    timeout = c_int32(-1 if timeout is None else timeout)
    return do_call_async(
        "indy_vdr_pool_submit_action",
        pool_handle,
        request_handle,
        nodes_p,
        timeout,
        return_type=lib_string,
        post_process=str,
    )


def pool_submit_request(
    pool_handle: PoolHandle, request_handle: RequestHandle
) -> asyncio.Future:
    """Publishes a prepared request message to the validator pool."""
    return do_call_async(
        "indy_vdr_pool_submit_request",
        pool_handle,
        request_handle,
        return_type=lib_string,
        post_process=str,
    )


def pool_close(pool_handle: PoolHandle):
    """Close and free a pool instance."""
    do_call("indy_vdr_pool_close", pool_handle)


def pool_get_transactions(pool_handle: PoolHandle) -> asyncio.Future:
    """Fetch the transactions for an opened pool instance."""
    return do_call_async(
        "indy_vdr_pool_get_transactions",
        pool_handle,
        return_type=lib_string,
        post_process=str,
    )


def pool_get_verifiers(pool_handle: PoolHandle) -> asyncio.Future:
    """Fetch the set of active verifiers for an opened pool instance."""
    return do_call_async(
        "indy_vdr_pool_get_verifiers",
        pool_handle,
        return_type=lib_string,
        post_process=str,
    )


def request_free(handle: RequestHandle):
    """Manually free a prepared request which won't be submitted."""
    do_call("indy_vdr_request_free", handle)


def request_get_body(handle: RequestHandle) -> str:
    """Get the canonical signature input for a prepared request."""
    body = lib_string()
    do_call("indy_vdr_request_get_body", handle, byref(body))
    return body.value.decode("utf-8")


def request_get_signature_input(handle: RequestHandle) -> bytes:
    """Get the canonical signature input for a prepared request."""
    sig_input = lib_string()
    do_call("indy_vdr_request_get_signature_input", handle, byref(sig_input))
    return sig_input.value


def request_set_endorser(handle: RequestHandle, endorser_did: str):
    """Set the endorser on a prepared request."""
    endorser_p = encode_str(endorser_did)
    do_call("indy_vdr_request_set_endorser", handle, endorser_p)


def request_set_signature(handle: RequestHandle, signature: bytes):
    """Set the signature on a prepared request."""
    sig_len = len(signature)
    do_call("indy_vdr_request_set_signature", handle, signature, sig_len)


def request_set_txn_author_agreement_acceptance(
    handle: RequestHandle, acceptance: Union[str, dict]
):
    """Set the transaction author agreement acceptance on a prepared request."""
    acceptance_p = (
        encode_str(acceptance)
        if isinstance(acceptance, str)
        else encode_json(acceptance)
    )
    do_call(
        "indy_vdr_request_set_txn_author_agreement_acceptance", handle, acceptance_p
    )


def set_config(config: dict):
    """Set the library configuration."""
    do_call("indy_vdr_set_config", encode_json(config))


def set_protocol_version(version: int):
    """Set the library protocol version."""
    do_call("indy_vdr_set_protocol_version", c_size_t(version))


def version() -> str:
    """Set the version of the installed indy_vdr library."""
    lib = get_library()
    lib.indy_vdr_version.restype = c_void_p
    return lib_string(lib.indy_vdr_version()).value.decode("utf-8")



"""Module setup."""

import runpy
from setuptools import find_packages, setup

PACKAGE_NAME = "indy_vdr"
version_meta = runpy.run_path("./{}/version.py".format(PACKAGE_NAME))
VERSION = version_meta["__version__"]

if __name__ == "__main__":
    setup(
        name=PACKAGE_NAME,
        version=VERSION,
        url="https://github.com/andrewwhitehead/indy-vdr",
        packages=find_packages(),
        include_package_data=True,
        package_data={
            "indy_vdr": ["indy_vdr.dll", "libindy_vdr.dylib", "libindy_vdr.so"]
        },
        python_requires=">=3.6.3",
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
        ],
    )



import logging
import os
import sys

from aiohttp import web
from .bindings import version
from .ledger import build_get_txn_request
from .pool import Pool, open_pool

LOGGER = logging.getLogger(__name__)
GENESIS_PATH: str = None
POOL: Pool = None
ROUTES = web.RouteTableDef()


@ROUTES.get("/genesis")
async def genesis(request):
    if not POOL:
        return web.Response(status=503)
    genesis = await POOL.get_transactions()
    LOGGER.info("got txns")
    return web.json_response(genesis)


@ROUTES.get("/txn")
async def txn(request):
    if not POOL:
        return web.Response(status=503)
    req = build_get_txn_request(None, None, 15)
    resp = await POOL.submit_request(req)
    return web.json_response(resp)


async def boot(app):
    global POOL
    POOL = await open_pool(transactions_path=GENESIS_PATH)


if __name__ == "__main__":
    LOGGER.info("indy-vdr version: %s", version())

    GENESIS_PATH = len(sys.argv) > 1 and sys.argv[1] or "genesis.txn"

    APP = web.Application()
    APP.add_routes(ROUTES)
    APP.on_startup.append(boot)
    LOGGER.info("Running webserver...")
    PORT = int(os.getenv("PORT", "8000"))
    web.run_app(APP, host="0.0.0.0", port=PORT)
